#ifndef PARAMETERS_H_
#define PARAMETERS_H_

#include "ap_int.h"
#include "ap_fixed.h"

#include "nnet_utils/nnet_helpers.h"
//hls-fpga-machine-learning insert includes
 
//hls-fpga-machine-learning insert weights

//hls-fpga-machine-learning insert layer-config

#endif
